/**
 * \file        rootsPlot.c
 * \brief       Roots plot functions.
 * \copyright   Copyright (C) 2006-2022 Ralf Hoppe <ralf.hoppe@dfcgen.de>
 */


/* INCLUDE FILES **************************************************************/

#include <errno.h>

#include "dfcProject.h"
#include "rootsPlot.h"
#include "cairoPlot.h"
#include "cfgSettings.h"


/* GLOBAL CONSTANT DEFINITIONS ************************************************/


/* GLOBAL VARIABLE DEFINITIONS ************************************************/


/* LOCAL TYPE DECLARATIONS ****************************************************/


/* LOCAL CONSTANT DEFINITIONS *************************************************/


/* LOCAL VARIABLE DEFINITIONS *************************************************/


static int plotRootIndex;                     /**< Index of next root to plot */
static GdkWindow *rootsPlotWindow = NULL; /**< window of roots plot \c GtkDrawingArea */


/* LOCAL MACRO DEFINITIONS ****************************************************/


/* LOCAL FUNCTION DECLARATIONS ************************************************/

static int rootsPlotInit (double start, double stop, void *pData);
static double plotRoot (double *px, void *pData);
static double returnZeroRoot (double *px, void *pData);
static void updateRootsMinMax (MATHPOLY *poly, gsl_complex *rmin, gsl_complex *rmax);
static int calcRoots (MATHPOLY *poly);
static gboolean rootsPlotDrawHandler (GtkWidget *widget, cairo_t *gc,
                                      gpointer user_data);
static double plotUnitCircleBottom (double *px, void *pData);
static double plotUnitCircleTop (double *px, void *pData);



/* LOCAL FUNCTION DEFINITIONS *************************************************/


/* FUNCTION *******************************************************************/
/** Initialization function for a roots plot. This function must be of type
 *  PLOT_FUNC_INIT from cairoPlot.h (see there).
 *
 *  \param start        The real-world start x-coordinate.
 *  \param stop         The real-world stop x-coordinate.
 *  \param pData        Pointer to polynomial (MATHPOLY) for which the roots
 *                      shall be plotted. If \p pData is NULL the function
 *                      returns -1.
 *
 *  \return  The function returns:
 *           - normally the degree of polynomial referenced by \p pData
 *           - a negative number, if \p pData is NULL.
 ******************************************************************************/
static int rootsPlotInit (double start, double stop, void *pData)
{
    MATHPOLY *poly = pData;

    if (poly == NULL)
    {
        return -1;
    } /* if */

    plotRootIndex = 0;

    return poly->degree;
} /* rootsPlotInit() */



/* FUNCTION *******************************************************************/
/** Plot function for a polynomial root. This function must be of type
 *  PLOT_FUNC_GET from cairoPlot.h (see there).
 *
 *  \param px           Pointer to buffer, which gets the x-coordinate (real
 *                      part) of the next root.
 *  \param pData        User application data pointer as passed to cairoPlot2d()
 *                      in element \a pData of structure PLOT_DIAG. Here it
 *                      must point to the polynomial (MATHPOLY) for which the
 *                      next root shall be plotted.
 *
 *  \return             The y-coordinate (imag. part) of the next root. If there
 *                      is no value at \p px, then it returns GSL_POSINF.
 ******************************************************************************/
static double plotRoot (double *px, void *pData)
{
    double y;
    MATHPOLY *poly = pData;

    if (plotRootIndex >= poly->degree)                        /* sanity check */
    {
        return GSL_POSINF;
    } /* if */

    *px = GSL_REAL (poly->root[plotRootIndex]);
    y = GSL_IMAG (poly->root[plotRootIndex]);
    ++plotRootIndex;

    return y;
} /* plotRoot() */



/* FUNCTION *******************************************************************/
/** Plot function for a polynomial root at 0,0. This function must be of type
 *  PLOT_FUNC_GET from cairoPlot.h (see there).
 *
 *  \param px           Pointer to buffer, which gets the x-coordinate (real
 *                      part) of the root.
 *  \param pData        User application data pointer as passed to cairoPlot2d()
 *                      in element \a pData of structure PLOT_DIAG. Here it
 *                      is unused, because return ZeroRoot() always returns 0,0.
 *
 *  \return             The y-coordinate (imag. part) of the next root (always 0).
 ******************************************************************************/
static double returnZeroRoot (double *px, void *pData)
{
    *px = 0.0;

    return 0.0;
} /* returnZeroRoot() */



/* FUNCTION *******************************************************************/
/** Plot function for a north-oriented semicircle with unity radius. This
 *  function must be of type PLOT_FUNC_GET from cairoPlot.h (see there).
 *
 *  \param px           Pointer to real-world x-coordinate.
 *  \param pData        User application data pointer as passed to cairoPlot2d()
 *                      in element \a pData of structure PLOT_DIAG. Here it is
 *                      unused.
 *
 *  \return             The y-coordinate \f$y=\sqrt{1-x^2}\f$ of the semicircle
 *                      associated with \p *px.
 ******************************************************************************/
static double plotUnitCircleTop (double *px, void *pData)
{
    double y;

    if (fabs (*px) <= 1.0)
    {
        y = sqrt (1 - *px * *px);
    } /* if */
    else
    {
        y = GSL_POSINF;
    } /* else */

    return y;
} /* plotUnitCircleTop() */



/* FUNCTION *******************************************************************/
/** Plot function for a south-oriented semicircle with unity radius. This
 *  function must be of type PLOT_FUNC_GET from cairoPlot.h (see there).
 *
 *  \param px           Pointer to real-world x-coordinate.
 *  \param pData        User application data pointer as passed to cairoPlot2d()
 *                      in element \a pData of structure PLOT_DIAG. Here it is
 *                      unused.
 *
 *  \return             The y-coordinate \f$y=-\sqrt{1-x^2}\f$ of the semicircle
 *                      associated with \p *px.
 ******************************************************************************/
static double plotUnitCircleBottom (double *px, void *pData)
{
    double y;

    if (fabs (*px) <= 1.0)
    {
        y = -sqrt (1 - *px * *px);
    } /* if */
    else
    {
        y = GSL_NEGINF;
    } /* else */

    return y;
} /* plotUnitCircleBottom() */



/* FUNCTION *******************************************************************/
/** Updates real/imag maximum and minimum values from roots positions.
 *
 *  \param poly         Polynomial which roots shall be compared with the
 *                      current maximum and minimum values.
 *  \param rmin         Current minimum value to be updated from \p poly->root.
 *  \param rmax         Current maximum value to be updated from \p poly->root.
 *
 ******************************************************************************/
static void updateRootsMinMax (MATHPOLY *poly, gsl_complex *rmin, gsl_complex *rmax)
{
    int i;

    for (i = 0; i < poly->degree; i++)
    {
        if (GSL_REAL (poly->root[i]) > GSL_REAL (*rmax))    /* max. real part */
        {
            GSL_SET_REAL (rmax, GSL_REAL (poly->root[i]));
        } /* if */

        if (GSL_IMAG (poly->root[i]) > GSL_IMAG (*rmax))       /* max. imag. part */
        {
            GSL_SET_IMAG (rmax, GSL_IMAG (poly->root[i]));
        } /* if */

        if (GSL_REAL (poly->root[i]) < GSL_REAL (*rmin))        /* min. real part */
        {
            GSL_SET_REAL (rmin, GSL_REAL (poly->root[i]));
        } /* if */

        if (GSL_IMAG (poly->root[i]) < GSL_IMAG (*rmin))       /* min. imag. part */
        {
            GSL_SET_IMAG (rmin, GSL_IMAG (poly->root[i]));
        } /* if */
    } /* for */
} /* updateRootsMinMax() */



/* FUNCTION *******************************************************************/
/** The function calculates the roots of a polynomial. The function assumes
 *  that all coefficients \f$c_i\f$ in \p poly are based on the following
 *  representation in \e Z domain:
    \f[
    p(z)=c_n z^{-n} + c_{n-1} z^{-(n-1)}+ \cdots + c_2 z^{-2} + c_1 z^{-1} + c_0
    \f]
 *  If we substitute \f$w=z^{-1}\f$ it gives
    \f[
    p(w)=c_n w^n + c_{n-1} w^{n-1}+ \cdots + c_2 w^2 + c_1 w + c_0
    \f]
 *  In that sense the function first calculates the roots \f$w_i\f$, then it
 *  transforms all back into \f$z_i\f$. So the linear factor presentation of
 *  \f$p(z)\f$ by the help of \f$z_i\f$ is equivalent to the polynomial
 *  presentation in \e Z domain.
 *
 *  \param poly         Pointer to polynomial that holds the coefficients in
 *                      \p poly->coeff and gets the roots in \p poly->roots.
 *
 *  \return             0 on success, else an error number.
 ******************************************************************************/
static int calcRoots (MATHPOLY *poly)
{
    int ret, i;

    ret = mathPolyCoeffs2Roots (poly);

    if (ret == 0)
    {
        for (i = 0; i < poly->degree; i++)
        {
            poly->root[i] = gsl_complex_inverse (poly->root[i]);
        } /* for */
    } /* if */

    return ret;
} /* calcRoots */



/* FUNCTION *******************************************************************/
/** Redraws a response widget in case an \e draw event is received.
 *
 *  \param widget       Pointer to \c GtkDrawingArea widget, which has to be
 *                      redrawn with a particular filter response.
 *  \param gc           \e Cairo graphics context.
 *  \param user_data    Pointer to response description (of type RESPONSE_WIN)
 *                      as supplied to function g_signal_connect() for \e draw
 *                      event.
 *  \return             TRUE if the callback handled the signal, and no further
 *                      handling is needed.
 ******************************************************************************/
static gboolean rootsPlotDrawHandler (GtkWidget *widget, cairo_t *gc,
                                      gpointer user_data)
{
    static const char realText[] = N_("Re(z)");
    static const char imagText[] = N_("Im(z)");

    static GdkRGBA circleColor[PLOT_COLOR_SIZE];
    static GdkRGBA rootsColor[PLOT_COLOR_SIZE];


    static PLOT_DIAG rplots[] =
    {
        {                                  /* stability boundary circle (top) */
            {realText}, {imagText}, NULL,                      /* x, y, pData */
            1.0, PLOT_STYLE_LINE_ONLY, 0,            /* thickness, style, num */
            NULL, NULL,                             /* progressFunc, initFunc */
            plotUnitCircleTop, NULL,                   /* sampleFunc, endFunc */
            circleColor,                                            /* colors */
        },
        {                               /* stability boundary circle (bottom) */
            {realText}, {imagText}, NULL,                      /* x, y, pData */
            1.0, PLOT_STYLE_LINE_ONLY, 0,            /* thickness, style, num */
            NULL, NULL,                             /* progressFunc, initFunc */
            plotUnitCircleBottom, NULL,                /* sampleFunc, endFunc */
            circleColor,                                            /* colors */
        },
        {                                                            /* zeros */
            {realText}, {imagText}, NULL,                      /* x, y, pData */
            2.0, PLOT_STYLE_CIRCLE_ONLY, 0,          /* thickness, style, num */
            NULL, rootsPlotInit, plotRoot, NULL, /* progress-, init-, sample-, endFunc */
            rootsColor,                                             /* colors */
        },
        {                                                            /* poles */
            {N_(realText)}, {imagText}, NULL,                  /* x, y, pData */
            2.0, PLOT_STYLE_CROSS_ONLY, 0,           /* thickness, style, num */
            NULL, rootsPlotInit, plotRoot, NULL, /* progress-, init-, sample-, endFunc */
            rootsColor,                                             /* colors */
        },
        {                                                            /* poles */
            {N_(realText)}, {imagText}, NULL,                  /* x, y, pData */
            2.0, PLOT_STYLE_CROSS_ONLY, 1,           /* thickness, style, num */
            NULL, NULL, returnZeroRoot, NULL, /* progress-, init-, sample-, endFunc */
            rootsColor,                                             /* colors */
        }
    };


    FLTCOEFF *pFilter = dfcPrjGetFilter ();

    if (pFilter != NULL)
    {
        int i;
        gsl_complex rmax, rmin, delta;
        GdkRGBA colorNormal, colorInactive;
        GtkStyleContext* styleContext;

        int points = 0;
        int numPlots = N_ELEMENTS (rplots);
        const CFG_DESKTOP* pPrefs = cfgGetDesktopPrefs ();
        GtkWidget* topWidget = gtk_widget_get_toplevel (widget);
        GdkWindow* topWindow = gtk_widget_get_window (topWidget);

        GdkCursor* cursor = gdk_cursor_new_from_name (
            gtk_widget_get_display (topWidget), GUI_CURSOR_IMAGE_WATCH);
        gdk_window_set_cursor (topWindow, cursor);
        g_object_unref (cursor);     /* free client side resource immediately */

        GSL_SET_COMPLEX (&rmax, 1, 1);        /* at least in interval [-1,+1] */
        GSL_SET_COMPLEX (&rmin, -1, -1);
        updateRootsMinMax (&pFilter->den, &rmin, &rmax);
        updateRootsMinMax (&pFilter->num, &rmin, &rmax);
        delta = gsl_complex_mul_real (gsl_complex_sub (rmax, rmin), 0.05);
        rmin = gsl_complex_sub (rmin, delta);                   /* 5% stretch */
        rmax = gsl_complex_add (rmax, delta);

        if (pFilter->factor != 0.0)                       /* roots are valid? */
        {
            rplots[2].pData = &pFilter->num; /* set the expected data pointer */
            rplots[3].pData = &pFilter->den;
        } /* if */
        else
        {
            rplots[2].pData = rplots[3].pData = NULL;
        } /* else */

        if (pFilter->num.degree == pFilter->den.degree)
        {                                     /* polynomials with same degree */
            --numPlots;        /* do not plot the last diagram (roots at 0,0) */
        } /* if */
        else
        {                                  /* note: polynomials are in z^{-1} */
            if (pFilter->den.degree > pFilter->num.degree)
            {
                rplots[4].style = PLOT_STYLE_CIRCLE_ONLY;
            } /* if */
            else
            {
                rplots[4].style = PLOT_STYLE_CROSS_ONLY;
            } /* else */
        } /* else */


        styleContext = gtk_widget_get_style_context (widget);
        gtk_style_context_get_color (styleContext, GTK_STATE_FLAG_NORMAL,
                                     &colorNormal);
        gtk_style_context_get_color (styleContext, GTK_STATE_FLAG_INSENSITIVE,
                                     &colorInactive);

        circleColor[PLOT_COLOR_LABELS] =
        circleColor[PLOT_COLOR_GRID] =
        circleColor[PLOT_COLOR_BOX] = 
        circleColor[PLOT_COLOR_AXIS_NAME] = colorInactive;

        circleColor[PLOT_COLOR_GRAPH] = colorInactive;

        rootsColor[PLOT_COLOR_GRID] =
        rootsColor[PLOT_COLOR_AXIS_NAME] =
        rootsColor[PLOT_COLOR_BOX] = colorInactive;

        rootsColor[PLOT_COLOR_LABELS] =
        rootsColor[PLOT_COLOR_GRAPH] = colorNormal;

        for (i = 0; i < numPlots; i++)
        {
            rplots[i].area.x = rplots[i].area.y = 0;
            rplots[i].area.width = gtk_widget_get_allocated_width (widget);
            rplots[i].area.height = gtk_widget_get_allocated_height (widget);
            rplots[i].x.prec = rplots[i].y.prec = pPrefs->outprec;
            rplots[i].x.start = GSL_REAL (rmin);
            rplots[i].x.stop = GSL_REAL (rmax);
            rplots[i].y.start = GSL_IMAG (rmin);
            rplots[i].y.stop = GSL_IMAG (rmax);

            points = cairoPlot2d (gc, &rplots[i]);
        } /* for */

        if (points < 0)
        {
        } /* if */

        gdk_window_set_cursor (topWindow, NULL);            /* restore cursor */
    } /* if */

    return TRUE;                                             /* stop emission */
} /* rootsPlotDrawHandler() */




/* EXPORTED FUNCTION DEFINITIONS **********************************************/


/* FUNCTION *******************************************************************/
/** Creates a \e GtkDrawingArea used for roots display.
 *
 *  \return             Pointer to \c GtkDrawingArea widget, which is used in
 *                      function rootsPlotUpdate() to draw polynomial roots.
 ******************************************************************************/
GtkWidget *rootsPlotCreate ()
{
    GtkWidget* widget = gtk_drawing_area_new ();

    rootsPlotWindow = gtk_widget_get_window (widget);
    g_signal_connect ((gpointer) widget, "draw",
                      G_CALLBACK (rootsPlotDrawHandler), NULL);
    return widget;
} /* rootsPlotCreate() */



/* FUNCTION *******************************************************************/
/** Re-calculates the roots of transfer \f$H(z)\f$ of a filter.
 *
 *  \param pFilter      Pointer to filter coefficients, for which the roots
 *                      shall be calculated. Set this to NULL, if the filter
 *                      (and therefore the roots too) is invalid.
 *
 ******************************************************************************/
void rootsPlotUpdate (FLTCOEFF *pFilter)
{
    if (pFilter != NULL)
    {
        if ((calcRoots (&pFilter->num) != 0) || (calcRoots (&pFilter->den) != 0))
        {
            pFilter->factor = 0.0;                       /* roots are invalid */
        } /* if */
        else
        {
            pFilter->factor = pFilter->num.coeff[0] / pFilter->den.coeff[0];
        } /* else */
    } /* if */

    rootsPlotRedraw ();

} /* rootsPlotUpdate() */



/* FUNCTION *******************************************************************/
/** Forces a asynchronous redraw of all transfer function roots.
 *
 ******************************************************************************/
void rootsPlotRedraw ()
{
    if (rootsPlotWindow != NULL)
    {
        gdk_window_invalidate_rect (rootsPlotWindow, NULL, FALSE);
    }
} /* rootsPlotRedraw() */


/******************************************************************************/
/* END OF FILE                                                                */
/******************************************************************************/
